
bl_info = {
    "name":         "OFP MLOD-P3D Exporter",
    "author":       "Lukas Pinkowski",
    "blender":      (2,6,5),
    "version":      (0,1,4),
    "location":     "File > Export",
    "description":  "Export to OFP MLOD-P3D",
    "category":     "Import-Export"
}

import ofp.mlod
import bpy
from bpy_extras.io_utils import ExportHelper
import mathutils as M
import math

def swap_vertex(v):
    return [v[0], v[2], v[1]]

def to_uv(p):
    return [(uv[0], 1.0 - uv[1]) for uv in p]

SETTING_EXCLUDE_LAST_LAYER = False

def export_p3d(output_path):
    mlod_p3d = ofp.mlod.P3DFile()
    mesh_objects = [o for o in bpy.data.objects if o.type == 'MESH']
    resolution = 1.0
    
    objects = [None for x in range(20)]
    for obj in mesh_objects:
        for i, flag in enumerate(obj.layers):
            if flag:
                objects[i] = obj
                break
    
    for obj in objects:
        if obj is None:
            continue
            
        print("Exporting LOD '%s'" % resolution)
        lod = mlod_p3d.create_LOD()
        mesh = obj.data
        mesh.calc_tessface()
        
        transform = obj.matrix_local.copy()
        location, rotation, scale = transform.decompose()
        rotation.invert()
        scale = M.Vector([scale.x * scale.x, scale.y * scale.y, scale.z * scale.z])
        
        sharp_edges = set()
        
        print("Exporting vertices")
        lod.vertices = []
        for vertex in mesh.vertices:
            v = rotation * M.Vector([scale.x * vertex.co.x, scale.y * vertex.co.y, scale.z * vertex.co.z])
            lod.vertices.append(ofp.mlod.Vertex((v.x, v.z, v.y)))
        print("Exporting normals")
        lod.normals = [ofp.mlod.Normal((v.normal.x, v.normal.z, v.normal.y)) for v in mesh.vertices]
        
        print("Exporting faces")
        lod.polygons = []
        for face in mesh.tessfaces:
            texture = "undefined"
            if face.material_index < len(mesh.materials) and mesh.materials[face.material_index] is not None:
                texture = mesh.materials[face.material_index].name
                
            #tmp = [x for x in mesh.tessface_uv_textures.active.data[face.index].uv]
            #tmp2 = [to_uv(uv) for uv in tmp]
            #print(tmp2)
            #uvs = [to_uv(uv) for uv in mesh.tessface_uv_textures.active.data[face.index].uv]
            try:
                uvs = to_uv(mesh.tessface_uv_textures.active.data[face.index].uv)
            except AttributeError:
                uvs = [(0.0, 0.0) for x in face.vertices]
            #print(uvs)
            f = ofp.mlod.Face(face.vertices, face.vertices, texture, uvs)
            lod.polygons.append(f)
        
        for edge in mesh.edges:
            if edge.use_edge_sharp:
                sharp_edges.update([(edge.vertices[0], edge.vertices[1])])
        lod.sharp_edges = list(set(sharp_edges))
        
        print("Exporting vertex groups")
        for group_index, vertex_group in enumerate(obj.vertex_groups):
            selection_name = vertex_group.name
            print("Exporting named selection '%s'" % selection_name)
            selection = ofp.mlod.SP3X_Selection([0.0 for x in lod.vertices], [0.0 for x in lod.polygons])
            lod.selections[selection_name] = selection
            
            for vertex_index, vertex in enumerate(mesh.vertices):
                for group in vertex.groups:
                    if group.group == group_index:
                        selection.vertices[vertex_index] = group.weight
                        break
    
        lod.resolution = resolution
        resolution += 1.0
    
    mlod_p3d.write_to_path(output_path)
    print("MLOD has been written successfully to '%s'" % output_path)


class P3D_MLOD_Exporter(bpy.types.Operator, ExportHelper):
    bl_idname = "export_mlod_p3d.fmt"
    bl_label = "OFP MLOD-P3D Exporter"
    filename_ext = ".p3d"
    
    filepath = bpy.props.StringProperty(subtype='FILE_PATH')
    
    def execute(self, context):
        export_p3d(self.filepath)
        return {'FINISHED'}
    
    def invoke(self, context, event):
        if not self.filepath:
            self.filepath = bpy.path.ensure_ext(bpy.data.filepath, ".p3d")
        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}
    
#output_path = "/home/lukas/tmp/DanCREWg.p3d"
#export_p3d(output_path)

def menu_func(self, context):
    self.layout.operator(P3D_MLOD_Exporter.bl_idname, text="OFP MLOD-P3D (.p3d)")

def register():
    bpy.utils.register_module(__name__)
    bpy.types.INFO_MT_file_export.append(menu_func)

def unregister():
    bpy.utils.unregister_module(__name__)
    bpy.types.INFO_MT_file_export.remove(menu_func)

if __name__ == "__main__":
    register()
    print("check main")
    