bl_info = {
    "name":         "OFP MLOD-P3D Importer",
    "author":       "Lukas Pinkowski",
    "blender":      (2,6,5),
    "version":      (0,1,1),
    "location":     "File > Import",
    "description":  "Import from OFP MLOD-P3D",
    "category":     "Import-Export"
}

import ofp.mlod
import bpy
from bpy_extras.io_utils import ImportHelper

def swap_vertex(v):
    return [v[0], v[2], v[1]]

def to_uv(p):
    return [(uv[0], 1.0 - uv[1]) for uv in p]

def import_p3d(input_path):
    mlod_p3d = ofp.mlod.P3DFile()
    mlod_p3d.read_from_path(input_path)
    
    basename = input_path.split('\\')[-1].split('/')[-1]
    
    for index, lod in enumerate(mlod_p3d.lods):
        name = "%s.%d" % (basename, index)
        mesh = bpy.data.meshes.new(name)
        
        vertices = [swap_vertex(x) for x in lod.vertices]
        faces = [face.vertex_indices for face in lod.polygons]
        uvs = [to_uv(face.uv) for face in lod.polygons]
        
        mesh.vertices.add(len(vertices))
        mesh.tessfaces.add(len(faces))
        mesh.tessface_uv_textures.new()
        
        for vi, vertex in enumerate(vertices):
            mesh.vertices[vi].co = vertex
        
        for fi, vertex_indices in enumerate(faces):
            if len(vertex_indices) == 3:
                mesh.tessfaces[fi].vertices = vertex_indices
            else:
                mesh.tessfaces[fi].vertices_raw = vertex_indices
        
        for face_index, uv in enumerate(uvs):
            for ui, value in enumerate(uv):
                #print("fi: %s, ui: %s" % (face_index, ui))
                tmp = mesh.tessface_uv_textures.active.data[face_index]
                #print("len(tmp.uv): %s" % len(tmp.uv))
                if len(tmp.uv) <= ui:
                    print("P3D_IMPORTER: WARNING: UV index(%s) doesn't fit into UV array(length: %s)" % (ui, len(tmp.uv)))
                else:
                    tmp.uv[ui] = value
                #mesh.tessface_uv_textures.active.data[face_index].uv[ui] = value
        
        mesh.update()
        mesh.validate()
        
        obj = bpy.data.objects.new(name, mesh)
        bpy.context.scene.objects.link(obj)
        if index != 0:
            obj.layers[min(19, index)] = True
            obj.layers[0] = False
        
        for name, selection in lod.selections.items():
            vertex_group = obj.vertex_groups.new(name)
            for index, weight in enumerate(selection.vertices):
                if weight > 0.0:
                    vertex_group.add([index], weight, 'ADD')
        
        # this routine creates extra vertex groups corresponding to textures in O2
        # Since there's a material limit in Blender, and P3Ds often use too many
        # textures at the same time, we use this hack.
        # Please see Sanctuary's Blender and O2 tutorial for an explanation
        # if you don't plan on using the UV-map from O2, you can feel safe to change
        # the next line to "if False:"
        SETTING_VERTEXGROUP_TEXTURE = True
        if SETTING_VERTEXGROUP_TEXTURE:
            texture_table = {}
            for face in lod.polygons:
                if face.texture:
                    if face.texture in texture_table:
                        texture_table[face.texture].extend(face.vertex_indices)
                    else:
                        texture_table[face.texture] = face.vertex_indices
            
            for key, value in texture_table.items():
                vertex_group = obj.vertex_groups.new("*%s" % key)
                vertex_group.add(list(set(value)), 1.0, 'ADD')
        
        
    bpy.context.scene.update()


class P3D_MLOD_Importer(bpy.types.Operator, ImportHelper):
    bl_idname = "import_mlod_p3d.fmt"
    bl_label = "OFP MLOD-P3D Importer"
    filename_ext = ".p3d"
    
    filepath = bpy.props.StringProperty(subtype='FILE_PATH')
    
    def execute(self, context):
        import_p3d(self.filepath)
        return {'FINISHED'}
    
    def invoke(self, context, event):
        if not self.filepath:
            self.filepath = bpy.path.ensure_ext(bpy.data.filepath, ".p3d")
        context.window_manager.fileselect_add(self)
        return {'RUNNING_MODAL'}

def menu_func(self, context):
    self.layout.operator(P3D_MLOD_Importer.bl_idname, text="OFP MLOD-P3D (.p3d)")

def register():
    bpy.utils.register_module(__name__)
    bpy.types.INFO_MT_file_import.append(menu_func)

def unregister():
    bpy.utils.unregister_module(__name__)
    bpy.types.INFO_MT_file_import.remove(menu_func)

if __name__ == "__main__":
    register()
    print("check main")



#input_path = "/home/lukas/tmp/DanCREWg.p3d"
#input_path = "/home/lukas/Projects/ofp/other/DanGuer/DanCREWg.p3d"
#import_p3d(input_path)
