import sys
import struct
import math
from ofp.resolutions import *

class Vertex:
    """
    A vertex is a point in space, most of the
    time a point of a triangle/quad.
    """
    def __init__(self, coords=[0.0,0.0,0.0], flags=0):
        self.xyz = list(coords)
        self.flags = flags
        
    def __getitem__(self, index):
        return self.xyz[index]
        
    def __setitem__(self, index, value):
        self.xyz[index] = value


class Normal:
    """
    Normals are used for lighting calculation
    """
    def __init__(self, coords=[0.0,1.0,0.0] ):
        self.xyz = list(coords)
        
    def __getitem__(self, index):
        return self.xyz[index]
        
    def __setitem__(self, index, value):
        self.xyz[index] = value


class Face:
    """
    Faces are either triangles or quads; they have indices into a vertex-
    and normal-array, specify the UV-coordinates, the texture and some
    special rendering flags.
    """
    def __init__(self, 
            vertex_indices=[], normal_indices=[], 
            texture="", uv=[], flags=0, material=""):
        """
        Creates a face (triangle or quad); vertex_indices and normal_indices are
        list of indices that point to a specific vertex and normal list.
        Texture is just the relative path to the textures.
        uv is a list of UV-pairs (the list should be either of length 3 or 4).
        flags is an unsigned integer representing some bit flags for 
        the face (for example 'Billboard'-flag)
        """
        self.vertex_indices = vertex_indices if vertex_indices else []
        self.normal_indices = normal_indices if normal_indices else []
        self.texture = texture # relative path
        self.material = material
        self.uv = uv if uv else [] # a list of UV: [ (0, 0), (1, 0), (0, 1), (1, 1) ]
        self.flags = flags
        
    def calc_area(self, vertices):
        """
        Calculates the area used by a specific face (with indices into the
        vertices array). The area is the half magnitude of the triangles
        normal vector (quads are split into two triangles).
        """
        def sub(a, b):
            return (b[0] - a[0], b[1] - a[1], b[2] - a[2])
            
        def cross(b, c):
            return (b[1] * c[2] - b[2] * c[1], 
                    b[2] * c[0] - b[0] * c[2], 
                    b[0] * c[1] - b[1] * c[0])
                    
        def magnitude(v):
            return math.sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2])
        
        v0 = vertices[self.vertex_indices[0]].xyz
        v1 = vertices[self.vertex_indices[1]].xyz
        v2 = vertices[self.vertex_indices[2]].xyz
        area = math.fabs(
            0.5 * magnitude(cross(sub(v0, v1), sub(v0, v2))))
        
        if len(self.vertex_indices) == 4:
            v3 = vertices[self.vertex_indices[3]].xyz
            area += math.fabs(
                0.5 * magnitude(cross(sub(v3, v1), sub(v3, v2))))
            
        return area
        
    def dup(self):
        return Face(self.vertex_indices, 
                    self.normal_indices, 
                    self.texture, self.uv, self.flags, self.material)

        
        
def weight_to_byte(weight):
    return 0 if weight < 0.004 else int(256 - 255*weight)

def byte_to_weight(byte):
    return 0.0 if byte == 0 else (256-byte)/255.0

def bool_to_selected(val):
    return 1 if val else 0
        
        
class SP3X_Selection:
    """
    Named selections give names along with weights to a set of vertices and faces.
    This is the SP3X (OFP) version
    """
    def __init__(self, vertices, faces):
        self.vertices_bytes = vertices # original byte version
        self.faces_bytes = faces # original byte version
        
        self.vertices = [byte_to_weight(x) for x in vertices] # vertex weights, float version (1.0 = 100%)
        self.faces = faces # list of True or False
        
    def to_P3DM(self):
        pass

class P3DM_Selection:
    """
    This is the P3DM (ArmA) version of Named Selections
    """
    def __init__(self, vertices, faces):
        self.vertices_bytes = vertices # original byte version
        self.faces_bytes = faces # original byte version
        
        self.vertices = [byte_to_weight(x) for x in vertices]
        self.faces = faces
        
    def to_SP3X(self):
        pass


class LOD:
    """
    The LOD holds all relevant information, like vertices, faces, normals.
    """
    def __init__(self):
        self.type = 'SP3X'
        self.vertices = []
        self.normals = []
        self.polygons = []
        self.mass = []
        self.properties = {} # named properties
        self.selections = {} # named selections (weights...)
        self.resolution = 0.0 # resolution
        self.sharp_edges = [] # list of vertex tuples...
        self.uvset = {} # UV sets in P3DM (ArmA MLOD)
        self.flags = 0
    
    def create_selection(self, name):
        weights = [0.0 for v in self.vertices]
        faceweights = [False for f in self.polygons]
        sel = SP3X_Selection([], faceweights)
        sel.vertices = weights
        self.selections[name] = sel
        return sel
        
        
    def set_SP3X(self):
        """
        Set LOD-type to SP3X
        """
        self.type = 'SP3X'
        
        
    def set_P3DM(self):
        """
        Set LOD-type to P3DM
        """
        self.type = 'P3DM'
        
        
    def convert_to_SP3X(self):
        """
        Convert this LOD to SP3X
        """
        if self.type == 'P3DM':
            self.type = 'SP3X'
        else:
            pass
            
            
    def convert_to_P3DM(self):
        """
        Convert this LOD to P3DM
        """
        if self.type == 'SP3X':
            self.type = 'P3DM'
        else:
            pass
        
        
class P3DParseError(Exception):
    pass


def get_z_string(file):
    """
    Read a 0-terminated string from file-like object;
    """
    ret = b""
    c = struct.unpack("c", file.read(1))[0]
    while c != b'\0':
        ret = ret + c
        c = struct.unpack("c", file.read(1))[0]
    return ret
    
        
class P3DFile:
    def __init__(self):
        self.lods = []
        self.OdolEx = False # if this P3D was made by OdolEx
        
    def create_LOD(self):
        """
        Create and add a new LOD
        """
        newlod = LOD()
        self.lods.append(newlod)
        return newlod
        
    def convert_to_OFP(self):
        for L in self.lods:
            L.convert_to_SP3X()
            
    def convert_to_Arma(self):
        for L in self.lods:
            L.convert_to_P3DM()
        
        
    def write_to_path(self, path):
        """
        Writes this P3D to a specific path.
        """
        with open(path, "wb") as file:
            self.write_to_file(file, False)
        
        
    def write_to_file(self, file, close=False, default_path="data\\"):
        """
        Writes this MLOD to a file-like object
        """
        file.write(b"MLOD")
        file.write(struct.pack("i", 0x00000101))
        file.write(struct.pack("i", len(self.lods)))
        for lod in self.lods:
            if lod.type == 'SP3X':
                self.write_SP3X(file, lod)
            elif lod.type == 'P3DM':
                self.write_P3DM(file, lod)
            else:
                assert(False)
            
        # at the very last: the default path:
        if default_path:
            file.write(struct.pack("32s", default_path.encode("cp1252")))
        
        if close:
            file.close()
    
    
    def write_SP3X(self, file, lod):
        """
        Writes OFP-style 'SP3X' Lod
        """
        file.write(b"SP3X")
        file.write(struct.pack("i", 0x0000001c))
        file.write(struct.pack("i", 0x00000099))
        
        # now vertex count, normal count, polycount and lod flags
        file.write(struct.pack("i", len(lod.vertices)))
        file.write(struct.pack("i", len(lod.normals)))
        file.write(struct.pack("i", len(lod.polygons)))
        file.write(struct.pack("i", lod.flags))
        
        # write vertex data
        for v in lod.vertices:
            file.write(struct.pack("fffI", v.xyz[0], v.xyz[1], v.xyz[2], v.flags))
            
        # write normal data
        for n in lod.normals:
            file.write(struct.pack("fff", n.xyz[0], n.xyz[1], n.xyz[2]))
            
        # write polygon data...
        # polygon structure is:
        # 32byte zero-terminated string
        # 4byte  triangle or quad (3 or 4)
        # 4 * 16byte  iiff (vertex+normal index, uv float)
        # 4byte face flags
        for p in lod.polygons:
            file.write(struct.pack(
                "32si", p.texture.encode("cp1252"), len(p.vertex_indices)))
            file.write(struct.pack("iiffiiffiiff", 
                p.vertex_indices[0], p.normal_indices[0], 
                p.uv[0][0], p.uv[0][1], 
                p.vertex_indices[1], p.normal_indices[1], 
                p.uv[1][0], p.uv[1][1], p.vertex_indices[2], p.normal_indices[2], 
                p.uv[2][0], p.uv[2][1]))
                
            if len(p.vertex_indices) == 3:
                file.write(struct.pack("iiff", 0, 0, 0.0, 0.0))
            else:
                file.write(struct.pack("iiff", 
                    p.vertex_indices[3], p.normal_indices[3], 
                    p.uv[3][0], p.uv[3][1]))
            file.write(struct.pack("I", p.flags))
                
        # now comes all the fun stuff, the tags...
        file.write(b"TAGG")
        
        # check for sharp edges:
        if lod.sharp_edges:
            file.write(struct.pack("64si", b"#SharpEdges#", 8 * len(lod.sharp_edges)))
            # the size is 2 * 4 * len(sharp_edges)
            for se in lod.sharp_edges:
                file.write(struct.pack("ii", se[0], se[1]))
            
        # now the properties
        if lod.properties:
            for key, value in lod.properties.items():
                file.write(struct.pack("64si", b"#Property#", 128))
                file.write(struct.pack("64s64s", 
                    key.encode("cp1252"), value.encode("cp1252")))
            
        # if we're in a geometry LOD, we need to write the mass values, too!
        if lod.resolution == GEOMETRY and lod.mass:
            file.write(struct.pack("64si", b"#Mass#", 4 * len(lod.mass)))
            for m in lod.mass:
                file.write(struct.pack("f", m))
            
        # now the selections
        for name, sel in lod.selections.items():
            assert(len(sel.vertices) == len(lod.vertices))
            assert(len(sel.faces) == len(lod.polygons))
            file.write(struct.pack("64si", 
                name.encode("cp1252"), 
                len(lod.vertices) + len(lod.polygons)))
                
            # the size is 1 * len(vertices) + 1 * len(polygons)
            for v in sel.vertices:
                file.write(struct.pack("B", weight_to_byte(v)))
            for p in sel.faces:
                file.write(struct.pack("B", bool_to_selected(p)))
                
            
        # end the Tags:
        file.write(struct.pack("64si", b"#EndOfFile#", 0))
        
        # last the Lod resolution:
        if isinstance(lod.resolution, float):
            file.write(struct.pack("f", lod.resolution))
        elif isinstance(lod.resolution, int):
            file.write(struct.pack('i', lod.resolution))
        else:
            raise P3DParseError('lod.resolution is neither float nor int!')
        
        
    def write_P3DM(self, file, lod):
        """
        Writes ArmA-style 'P3DM' Lod
        """
        file.write(b"P3DM")
        file.write(struct.pack("i", 0x0000001c))
        file.write(struct.pack("i", 0x00000100))
        
        # now vertex count, normal count, polycount and lod flags
        file.write(struct.pack("i", len(lod.vertices)))
        file.write(struct.pack("i", len(lod.normals)))
        file.write(struct.pack("i", len(lod.polygons)))
        file.write(struct.pack("i", 0))
        
        # write vertex data
        for v in lod.vertices:
            file.write(struct.pack("fffI", v.xyz[0], v.xyz[1], v.xyz[2], v.flags))
            
        # write normal data
        for n in lod.normals:
            file.write(struct.pack("fff", n.xyz[0], n.xyz[1], n.xyz[2]))
            
        # write polygon data...
        # polygon structure is:
        # 4byte  triangle or quad (3 or 4)
        # 4 * 16byte  iiff (vertex+normal index, uv float)
        # 4byte face flags
        # 2 * zero-terminated string (texture and material)
        for p in lod.polygons:
            file.write(struct.pack('i', len(p.vertex_indices)))
            file.write(struct.pack("iiffiiffiiff", 
                    p.vertex_indices[0], p.normal_indices[0], 
                    p.uv[0][0], p.uv[0][1], 
                    p.vertex_indices[1], p.normal_indices[1], 
                    p.uv[1][0], p.uv[1][1], 
                    p.vertex_indices[2], p.normal_indices[2], 
                    p.uv[2][0], p.uv[2][1]))
            if len(p.vertex_indices) == 3:
                file.write(struct.pack("iiff", 0, 0, 0.0, 0.0))
            else:
                file.write(struct.pack("iiff", 
                    p.vertex_indices[3], p.normal_indices[3], 
                    p.uv[3][0], p.uv[3][1]))
            file.write(struct.pack("I", p.flags))
            file.write(p.texture + '\x00' if p.texture else '\x00')
            file.write(p.material + '\x00' if p.material else '\x00')
                
        # now comes all the fun stuff, the tags...
        file.write(b"TAGG")
        
        # check for sharp edges:
        if lod.sharp_edges:
            file.write(struct.pack("B", 1))
            file.write(b'#SharpEdges\x00')
            file.write(struct.pack('i', 8*len(lod.sharp_edges)))
            # the size is 2 * 4 * len(sharp_edges)
            for se in lod.sharp_edges:
                file.write(struct.pack("ii", se[0], se[1]))
            
        # now the properties
        if lod.properties:
            for key, value in lod.properties.iteritems():
                file.write(struct.pack("B", 1))
                file.write(b'#Property#\x00')
                file.write(struct.pack("i", 128))
                file.write(struct.pack("64s64s", key, value))
            
        # if we're in a geometry LOD, we need to write the mass values, too!
        if lod.resolution == LOD_GEOMETRY and lod.mass:
            file.write(struct.pack("B", 1))
            file.write(b'#Mass#\x00')
            file.write(4 * len(lod.mass))
            for m in lod.mass:
                file.write( struct.pack("f", m) )
            
        # now the selections
        for name, sel in lod.selections.iteritems():
            file.write(struct.pack("B", 1))
            file.write((name + '\x00').encode("cp1252"))
            file.write(struct.pack('i', 
                len(lod.vertices) + len(lod.polygons)))
            for v in lod.vertex_selected:
                file.write(struct.pack("B", weight_to_byte(v)))
            for p in lod.face_selected:
                file.write(struct.pack("B", bool_to_selected(p)))
                
        # TODO: implement UV set
        if lod.uvset:
            file.write(lod.uvset)
                
        # end the Tags:
        file.write(struct.pack("B", 1))
        file.write(b'#EndOfFile#\x00')
        file.write(struct.pack('i', 0))
        
        # last the Lod resolution:
        if isinstance(lod.resolution, float):
            file.write(struct.pack("f", lod.resolution))
        elif isinstance(lod.resolution, int):
            file.write(struct.pack('i', lod.resolution))
        else:
            raise P3DParseError('lod.resolution is neither float nor int!')
        
    
    
    def read_from_path(self, path):
        """
        Read a P3D from a path (calls then readFromFile)
        """
        with open(path, "rb") as file:
            self.read_from_file(file, False)
    
        
    def read_from_file(self, file, close=False):
        """
        Read a P3D from a file-like object (if close is set, then 
        file-object is closed).
        """
        fourcc = struct.unpack("4s", file.read(4))[0]
        if fourcc == b'ODOL':
            raise P3DParseError("File is in ODOL-format (ODOL to MLOD not implemented yet)!")
        elif fourcc != b"MLOD":
            raise P3DParseError("File is not a MLOD file!")
            
        vers = struct.unpack("i", file.read(4))[0]
        if vers == 0x3f8ccccd:
            print("P3D was converted by ODOL Explorer, Enabled OdolEx compatibility mode.")
            self.OdolEx = True
        elif vers != 0x00000101:
            raise P3DParseError("Only MLOD version 0x00000101 supported, found: 0x%x" % vers)
            
        lod_count = struct.unpack("i", file.read(4))[0]
        for n in range(lod_count):
            lod_tag = struct.unpack("4s", file.read(4))[0]
            if lod_tag == b'SP3X':
                self.read_SP3X(file)
            elif lod_tag == b'P3DM':
                self.read_P3DM(file)
            else:
                raise P3DParseError("fourcc for LODs needs to be 'SP3X' or 'P3DM', not '%s'" % lod_tag)
            
        if close:
            file.close()
            
            
    def read_SP3X(self, file):
        if struct.unpack("i", file.read(4))[0] != 0x0000001c:
            raise P3DParseError("Wrong SP3X major version number %d" % vers)
        
        if struct.unpack("i", file.read(4))[0] != 0x00000099:
            raise P3DParseError("Wrong SP3X minor version number %d" % vers)
            
        newlod = self.create_LOD()
        newlod.set_SP3X()
        
        vertex_count = struct.unpack("i", file.read(4))[0]
        normal_count = struct.unpack("i", file.read(4))[0]
        poly_count = struct.unpack("i", file.read(4))[0]
        newlod.flags = struct.unpack("i", file.read(4))[0]
        
        for v in range(vertex_count):
            fvertex = struct.unpack("3fI", file.read(16))
            newlod.vertices.append(Vertex(fvertex[:3], fvertex[3]))
            
        for v in range(normal_count):
            fnormal = struct.unpack("3f", file.read(12))
            newlod.normals.append(Normal(fnormal))
            
        fmt = "32siiiffiiffiiffiiffI"
        fmts = struct.calcsize(fmt)
        
        for v in range(poly_count):
            fpoly = struct.unpack(fmt, file.read(fmts))
            
            # everything after the terminating 0 needs to be
            # stripped from the string (there's not guarantee
            # it's all filled with 0s).
            path = fpoly[0].decode("cp1252")
            path = path[:path.find('\x00')]
            
            num_vertices = fpoly[1]
            if num_vertices == 3:
                vIndex = [fpoly[2], fpoly[6], fpoly[10]]
                nIndex = [fpoly[3], fpoly[7], fpoly[11]]
                uv = [(fpoly[4], fpoly[5]), 
                      (fpoly[8],fpoly[9]), 
                      (fpoly[12],fpoly[13])]
            else:
                vIndex = [fpoly[2], fpoly[6], fpoly[10], fpoly[14]]
                nIndex = [fpoly[3], fpoly[7], fpoly[11], fpoly[15]]
                uv = [(fpoly[4], fpoly[5]), 
                      (fpoly[8],fpoly[9]), 
                      (fpoly[12],fpoly[13]), 
                      (fpoly[16],fpoly[17])]
                
            flags = fpoly[18]
            new_face = Face(vIndex, nIndex, path.lower(), uv, flags)
            newlod.polygons.append(new_face)
            
        # now the taggs...
        if struct.unpack("4s", file.read(4))[0] != b"TAGG":
            raise P3DParseError("Missing 'TAGG' fourcc after polygon data, found: '%s'" % magic)
        
        while True:
            taggname = struct.unpack("64s", file.read(64))[0].decode("cp1252")
            taggname = taggname[:taggname.find('\x00')]
            taggsize = struct.unpack("i", file.read(4))[0]
            
            if taggname == "#EndOfFile#":
                break
            elif taggname == "#SharpEdges#":
                for i in range(taggsize // 8):
                    newlod.sharp_edges.append(struct.unpack("ii", file.read(8)))
            elif taggname == "#Property#":
                for v in range(taggsize // 128):
                    key, value = struct.unpack("64s64s", file.read(128))
                    key = key.decode("cp1252")
                    value = value.decode("cp1252")
                    newlod.properties[key[:key.find('\x00')]] = value[:value.find('\x00')]
                    
            elif taggname == "#MaterialIndex#":
                file.seek(taggsize,1)
            elif taggname == "#Animation#":
                file.seek(taggsize,1)
            elif taggname == "#Mass#":
                if len(newlod.vertices) != taggsize / 4 and not self.OdolEx:
                    raise P3DParseError("Taggsize expected to be %d, not %d" % (len(newlod.vertices), taggsize))
                for v in range(len(newlod.vertices)):
                    newlod.mass.append(struct.unpack("f", file.read(4))[0])
            elif taggname == "#Selected#":
                file.seek(taggsize,1)
            elif taggname == "#Hide#":
                file.seek(taggsize,1)
            elif taggname == "#Lock#":
                file.seek(taggsize,1)
            elif taggname == "#UVSet#":
                file.seek(taggsize,1)
            elif taggname[0] == '#' and taggname[-1] == '#':
                print("Skipped unknown tagg: %s" % taggname)
                file.seek(taggsize,1)
            else:
                if (len(newlod.vertices) + len(newlod.polygons)) != taggsize and not self.OdolEx:
                    raise P3DParseError("Taggsize expected to be %d, not %d" % (len(newlod.vertices) + len(newlod.polygons), taggsize))
                vertices = [struct.unpack("B", file.read(1))[0] for X in range(len(newlod.vertices))]
                faces = [struct.unpack("B", file.read(1))[0] for X in range(len(newlod.polygons))]
                newlod.selections[taggname.lower()] = SP3X_Selection(vertices, faces)
                
        res_string = file.read(4)
        float_value = struct.unpack('f', res_string)[0]
        int_value = struct.unpack('i', res_string)[0]
        
        newlod.resolution = float_value if float_value < 1.0e3 else int_value
            
            
    def readP3DM(self, file):
        if struct.unpack("i", file.read(4))[0] != 0x0000001c:
            raise P3DParseError("Wrong P3DM major version number %d" % vers)
        
        if struct.unpack("i", file.read(4))[0] != 0x00000100: # 255
            raise P3DParseError("Wrong P3DM minor version number %d" % vers)
            
        newlod = self.createLOD()
        newlod.setP3DM()
        
        vertexCount = struct.unpack("i", file.read(4))[0]
        normalCount = struct.unpack("i", file.read(4))[0]
        polyCount = struct.unpack("i", file.read(4))[0]
        flags = struct.unpack("i", file.read(4))[0]
        
        for v in xrange(vertexCount):
            fvertex = struct.unpack("3fI", file.read(16))
            newlod.vertices.append( Vert(fvertex[:3], fvertex[3]) )
            
        for v in xrange(normalCount):
            fnormal = struct.unpack("3f", file.read(12))
            newlod.normals.append( Normal(fnormal) )
            
        fmt = "iiiffiiffiiffiiffI"
        fmts = struct.calcsize(fmt)
        
        for v in range(polyCount):
            fpoly = struct.unpack(fmt, file.read(fmts))
            
            numvert = fpoly[0]
            vIndex = [ fpoly[1], fpoly[5], fpoly[9] ]
            nIndex = [ fpoly[2], fpoly[6], fpoly[10] ]
            uv = [ (fpoly[3], fpoly[4]), (fpoly[7],fpoly[8]), (fpoly[11],fpoly[12]) ]
                
            if numvert == 4:
                vIndex.append(fpoly[13])
                nIndex.append(fpoly[14])
                uv.append((fpoly[15],fpoly[16]))
                
            flags = fpoly[17]
            
            path = get_z_string(file)
            material = get_z_string(file)
            
            newface = Face( vIndex, nIndex, path.lower(), uv, flags )
            newface.material = material
            
            newlod.polygons.append(newface)
            
        # now the taggs...
        magic = struct.unpack("4s", file.read(4))[0]
        if magic != b"TAGG":
            raise P3DParseError("Missing 'TAGG' fourcc after polygon data, found: '%s'" % magic)
            
        while True:
            taggbyte = struct.unpack("B", file.read(1))[0] # what is this supposed to be?
            taggname = get_z_string(file)
            taggsize = struct.unpack("i", file.read(4))[0]
            
            if( taggname == "#EndOfFile#" ):
                break
                
            elif( taggname == "#SharpEdges#" ):
                for i in xrange(taggsize / 8):
                    newlod.sharp_edges.append(struct.unpack("ii", file.read(8)))
                
            elif( taggname == "#Property#" ):
                for v in xrange(taggsize / 128):
                    key, value = struct.unpack("64s64s", file.read(128))
                    newlod.properties[key[:key.find('\x00')]] = value[:value.find('\x00')]
                    
            elif( taggname == "#MaterialIndex#" ):
                file.seek(taggsize, 1)
                
            elif( taggname == "#Animation#" ):
                file.seek(taggsize,1)
                
            elif( taggname == "#Mass#" ):
                for v in xrange(taggsize / 4):
                    newlod.mass.append( struct.unpack("f", file.read(4))[0] )
                
            elif taggname in ['#Selected#', '#Hide#', '#Lock#']:
                file.seek(taggsize, 1)
                
            elif( taggname == "#UVSet#" ):
                # read in the UVset, correct it later!
                newlod.uvset = file.read(taggsize)
                
            elif( taggname[0] == '#' and taggname[-1] == '#' ):
                print("Skipped unknown tagg: %s" % taggname)
                file.seek(taggsize, 1)
                
            else:
                #newselection = P3DM_Selection()
                if (len(newlod.vertices) + len(newlod.polygons)) != taggsize:
                    raise P3DParseError("Taggsize expected to be %d, not %d" % (len(newlod.vertices), taggsize))
                vertices = [struct.unpack("B", file.read(1))[0] for X in xrange(len(newlod.vertices))]
                faces = [struct.unpack("B", file.read(1))[0] for X in xrange(len(newlod.polygons))]
                newlod.selections[taggname.lower()] = P3DM_Selection(vertices, faces)
                
        newlod.resolution = struct.unpack("f", file.read(4))[0]
            
        
